package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.util.ChModPermissionsUtil;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.Job;
import be.iminds.ilabt.jfed.highlevel.jobs.SlicedState;
import be.iminds.ilabt.jfed.highlevel.jobs.State;
import be.iminds.ilabt.jfed.highlevel.jobs.StateSlice;
import be.iminds.ilabt.jfed.highlevel.util.ByteArraySshjFile;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.highlevel.util.ProxySocketFactoryProvider;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.rspec.model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.DistributeSshKeypair;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.BestNodeLoginFinder;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.sftp.*;
import net.schmizz.sshj.transport.TransportException;
import net.schmizz.sshj.transport.verification.PromiscuousVerifier;
import net.schmizz.sshj.userauth.UserAuthException;
import net.schmizz.sshj.userauth.keyprovider.KeyPairWrapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.net.SocketFactory;
import java.io.IOException;
import java.security.KeyPair;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

/**
 * ExperimentSpecificationServicesDirsAndUploadsState does 3 things:
 *   - Setup SSH connections to nodes, for the dirs, upload, and execute steps
 *   - Do the "dir" step of the ExperimentSpecification
 *   - Do the "upload" step of the ExperimentSpecification
 *
 * Note that these 2 steps can be combined here easily, as there are no inter node dependencies.
 * For the execute step, things are different. See ExperimentSpecificationServicesSingleExecuteState
 *
 * Additionally, this class can be called to close all SSH connections. (typically after all executes)
 *   See TODO
 */
public class ExperimentSpecificationServicesDirsAndUploadsState extends SlicedState<ExperimentSpecificationServicesDirsAndUploadsState.ExperimentSpecificationServiceDirsAndUploadsStateSlice> {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentSpecificationServicesDirsAndUploadsState.class);
    public static final int MAX_EXECUTE_FINISHED_FILE_DETECTION_TRIES = 60;
    public static final int TIME_BETWEEN_FINISHED_FILE_DETECTION_TRIES_MS = 10_000;

    @Nonnull
    private final Job<?> job;
    @Nonnull
    private final Experiment experiment;
    @Nonnull
    private final GeniUserProvider geniUserProvider;
    @Nonnull
    private final ProxySocketFactoryProvider proxySocketFactoryProvider;

    @Nonnull
    private final ExperimentSpecification experimentSpecification;
    @Nonnull
    private final ESpecBundle experimentSpecificationBundle;
    @Nonnull
    private final ESpecLogger eSpecLogger;

    @Nonnull
    private final KeyPair loginKeyPair;


    @Nonnull
    private final List<ExperimentSpecificationServiceDirsAndUploadsStateSlice> slices;

    @Nullable
    private final BestNodeLoginFinder bestNodeLoginFinder;

    @Nonnull
    private final ConcurrentHashMap<String, SSHClient> sshClientByClientId = new ConcurrentHashMap<>();
    @Nonnull
    private final ConcurrentHashMap<UploadLikeSpec, String> remotePathByUploadLikeSpec = new ConcurrentHashMap<>();

    protected ExperimentSpecificationServicesDirsAndUploadsState(@Nonnull Job<?> job,
                                                                 @Nonnull GeniUserProvider geniUserProvider,
                                                                 @Nonnull ProxySocketFactoryProvider proxySocketFactoryProvider) {
        super("Execute ExperimentSpecification Dir And Upload Steps");
        this.job = job;
        this.experiment = job.getExperiment();
        this.eSpecLogger = experiment.getExperimentSpecificationLogger();
        this.geniUserProvider = geniUserProvider;
        this.proxySocketFactoryProvider = proxySocketFactoryProvider;

        if (experiment.getExperimentSpecification() == null)
            throw new RuntimeException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification");
        if (experiment.getExperimentSpecificationBundle() == null)
            throw new RuntimeException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification (bundle)");

        this.experimentSpecification = experiment.getExperimentSpecification();
        this.experimentSpecificationBundle = experiment.getExperimentSpecificationBundle();

        assert experiment.getSliceOrNull() != null;

        if (experiment.getSliceOrNull() == null)
            throw new IllegalStateException("Experiment should have a slice");

        assert experiment.getKeypairs() != null;

        String distributedSshKeysUsername = null;
        KeyPair loginKeyPair = null;

        for (Map.Entry<DistributeSshKeypair, KeyPair> e : experiment.getKeypairs().entrySet()) {
            if (e.getValue() != null) {
                //distributedSshKeysUsername may be null, but that's ok, that just means the logged in users account needs to be used.
                distributedSshKeysUsername = e.getKey().getUser() == null ? null : e.getKey().getUser().getResourceName();

                loginKeyPair = e.getValue();
                break;
            }
        }

        if (loginKeyPair == null)
            loginKeyPair = new KeyPair(geniUserProvider.getLoggedInGeniUser().getPublicKey(),
                    geniUserProvider.getLoggedInGeniUser().getPrivateKey());

        this.loginKeyPair = loginKeyPair;


        this.slices = new ArrayList<>();

        RequestRspecSource requestRspecSource = experiment.getSliceOrNull().getRequestRspec();
        FXModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec(ModelRspecType.FX);
        assert model != null;

        if (model == null) {
            LOG.warn("ExperimentSpecificationServicesDirsAndUploadsState will do nothing because there is no usable request RSpec");
            bestNodeLoginFinder = null;
            return;
        }

        for (FXRspecNode node : model.getNodes()) {
            String nodeId = node.getClientId();

            if (nodeId != null) {
                //Note: Since this needs to setup SSH connections for execute as well, this will also run on nodes with only an execute step.

                boolean neededOnNode = experimentSpecification.getDirs().stream().anyMatch(s -> s.getNodes() == null || s.getNodes().contains(nodeId));
                neededOnNode |= experimentSpecification.getExecutes().stream().anyMatch(s -> s.getNodes() == null || s.getNodes().contains(nodeId));
                neededOnNode |= experimentSpecification.getUploads().stream().anyMatch(s -> s.getNodes() == null || s.getNodes().contains(nodeId));

                if (neededOnNode) {
                    slices.add(new ExperimentSpecificationServiceDirsAndUploadsStateSlice(job, node));
                }
            }
        }

        ManifestRspecSource manifestRspecSource = experiment.getSliceOrNull().getManifestRspec();
        bestNodeLoginFinder = manifestRspecSource == null ||
                manifestRspecSource.getStringRspec() == null ? null :
                new BestNodeLoginFinder(manifestRspecSource.getStringRspec(),
                        distributedSshKeysUsername, this.geniUserProvider.getLoggedInGeniUser(),
                        new BestNodeLoginFinder.Feedback() {
                            @Override
                            public void info(String s) {
                                LOG.info("ExperimentSpecificationServicesDirsAndUploadsState: " + s);
                            }

                            @Override
                            public void error(String s) {
                                LOG.error("ExperimentSpecificationServicesDirsAndUploadsState: " + s);
                            }

                            @Override
                            public void debug(String s) {
                                LOG.debug("ExperimentSpecificationServicesDirsAndUploadsState: " + s);
                            }
                        });
    }

    @Nonnull
    @Override
    protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
        ExperimentTaskStatus res = super.executeState(job);
        eSpecLogger.firePreExecute();
        return res;
    }

    private void closeAllSshConnections() {
        LOG.debug("closeAllSshConnections() called");
        for (SSHClient ssh : sshClientByClientId.values()) {
            try {
                if (ssh != null)
                    ssh.close();
            } catch (IOException ignored) {
                //probably, connection is already closed. We don't really need to worry about this.
                LOG.debug("Will ignore error while closing SSH connection: " + ignored.getMessage());
            }
        }
        //make sure it isn't reused
        sshClientByClientId.clear();
    }

    @Override
    public Collection<ExperimentSpecificationServiceDirsAndUploadsStateSlice> getSlices() {
        return slices;
    }

    @Nullable
    public SSHClient getSshClient(@Nonnull String clientId) {
        return sshClientByClientId.get(clientId);
    }

    @Nullable
    public String getRemotePath(@Nonnull UploadLikeSpec uploadLikeSpec) {
        return remotePathByUploadLikeSpec.get(uploadLikeSpec);
    }


    public class AssureAllSshClosedStateSlice extends State {
        protected AssureAllSshClosedStateSlice() {
            super("Close All SSH Connections");
        }

        @Nonnull
        @Override
        protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
            closeAllSshConnections();
            return ExperimentTaskStatus.SUCCESS;
        }
    }

    public AssureAllSshClosedStateSlice getAssureAllSshClosedStateSlice() {
        return new AssureAllSshClosedStateSlice();
    }

    public class ExperimentSpecificationServiceDirsAndUploadsStateSlice extends StateSlice {
        @Nonnull
        protected final FXRspecNode node;
        @Nonnull
        protected final String clientId;
        @Nullable
        private SSHClient ssh;

        @Nullable
        private ExternalFileUtil externalFileUtil;

        public ExperimentSpecificationServiceDirsAndUploadsStateSlice(@Nonnull Job<?> job,
                                                                      @Nonnull FXRspecNode node) {
            super(job);
            this.node = node;
            if (node.getClientId() == null)
                throw new IllegalStateException("Only nodes with a client_id can be used in an ExperimentSpecification");
            this.clientId = node.getClientId();
        }


        private String dirPathToAbsolutePath(String dirPath) {
            if (externalFileUtil == null)
                throw new IllegalStateException("externalFileUtil == null");

            //3 cases for path:
            //  - absolute path (starts with /)   example: /work/data/ / /work /work/test/data
            //  - home path (starts with ~)       exammple: ~/data/  ~/   ~   ~/test/data
            //  - null -> home path               -> this is handled by the parser, so no need to check this here
            //TODO    empty path string or path not starting with / or ~  ->  error during parsing

            boolean absolutePath = dirPath.startsWith("/");
            assert absolutePath || dirPath.startsWith("~");

            if (absolutePath)
                return assureEndWithSlash(dirPath);

            String homeDir = assureEndWithSlash(externalFileUtil.getHomeDir());

            return assureEndWithSlash(homeDir + (dirPath.length() > 1 ? dirPath.substring(1) : ""));
        }

        @Override
        public String getName() {
            return "ExperimentSpecification steps for node " + node.getClientId();
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            final SetupSshConnectionState setupSshConnectionState = new SetupSshConnectionState();
            setAndRunState(setupSshConnectionState);

            if (setupSshConnectionState.getStatus() != ExperimentTaskStatus.SUCCESS)
                return ExperimentTaskStatus.FAILED;

            eSpecLogger.firePreDir();
            //TODO: only start when needed
            final SetupDirsState setupDirState = new SetupDirsState();
            setAndRunState(setupDirState);
            eSpecLogger.firePostDirAll(setupDirState.getStatus() == ExperimentTaskStatus.SUCCESS);

            if (setupDirState.getStatus() != ExperimentTaskStatus.SUCCESS)
                return ExperimentTaskStatus.FAILED;


            eSpecLogger.firePreUpload();
            final UploadState uploadState = new UploadState();
            try {
                setAndRunState(uploadState);
            } finally {
                eSpecLogger.firePostUploadAll(uploadState.getStatus() == ExperimentTaskStatus.SUCCESS);
            }

            if (uploadState.getStatus() != ExperimentTaskStatus.SUCCESS)
                return ExperimentTaskStatus.FAILED;

            return ExperimentTaskStatus.SUCCESS;
        }

        public class SetupSshConnectionState extends State {

            protected SetupSshConnectionState() {
                super("Setting up SSH Connection to '" + node.getClientId() + "'");
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                //get login-service associated to the user we are distributing the key to
                BasicStringRspec.LoginService loginService =
                        bestNodeLoginFinder == null ?
                                null :
                                bestNodeLoginFinder.findBestLogin(node.getUniqueId());

                if (loginService == null) {
                    LOG.warn("Cannot setup connection to {} as no login-service is available", node.getClientId());
                    return ExperimentTaskStatus.FAILED;
                }

                ssh = new SSHClient();
                ssh.addHostKeyVerifier(new PromiscuousVerifier());

                try {
                    SocketFactory socketFactory = proxySocketFactoryProvider.createProxySocketFactory(loginService.getHostname(), loginService.getPort());
                    if (socketFactory != null) {
                        LOG.debug("Using proxy for ExperimentSpecification");
                        ssh.setSocketFactory(socketFactory);
                    } else {
                        LOG.debug("No proxy needed for ExperimentSpecification");
                    }
                } catch (IOException ex) {
                    LOG.error("Error while creating ProxySocketFactory. Trying without now.", ex);
                }

                try {
                    ssh.connect(loginService.getHostname(), loginService.getPort());
                } catch (IOException e) {
                    updateMessage("Error while connecting to " + loginService.getHostname() + ": " + e.getMessage());
                    LOG.error("Error while connecting to {}: {}", loginService.getHostname(), e.getMessage(), e);
                    return ExperimentTaskStatus.FAILED;
                }
                try {
                    //note: we must use the distributed ssh keys to login: in speaksfor case, there is no other way to login
                    ssh.authPublickey(loginService.getUsername(), new KeyPairWrapper(loginKeyPair));

                    externalFileUtil = new ExternalFileUtil(ssh);
                } catch (UserAuthException e) {
                    updateMessage("Could not authenticate: " + e.getMessage());
                    LOG.error("Authentication failed on {}: {}", node.getClientId(), e.getMessage(), e);
                    return ExperimentTaskStatus.FAILED;
                } catch (TransportException e) {
                    updateMessage("Error while authenticating: " + e.getMessage());
                    LOG.error("Error while authenticating on {}: {}", node.getClientId(), e.getMessage(), e);
                    return ExperimentTaskStatus.FAILED;
                }

                //only store ssh clients if the connection succeeded
                sshClientByClientId.put(clientId, ssh);

                return ExperimentTaskStatus.SUCCESS;

            }
        }

        public class SetupDirsState extends State {
            protected SetupDirsState() {
                super("Checking (and possible creating) dirs on " + node.getClientId());
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                if (ssh == null)
                    throw new IllegalStateException("ssh == null");
                if (externalFileUtil == null)
                    throw new IllegalStateException("externalFileUtil == null");

                List<DirSpec> dirs = new ArrayList<>(experimentSpecification.getDirs());

                for (DirSpec dirSpec : dirs) {
                    if (dirSpec.getNodes() == null || dirSpec.getNodes().contains(clientId)) {
                        String remotePath = dirPathToAbsolutePath(dirSpec.getPath());

                        int targetDirMask = ChModPermissionsUtil.toMask(dirSpec.getPermissions());
                        FileAttributes targetDirAtts = new FileAttributes.Builder()
                                .withPermissions(targetDirMask)
                                .withType(FileMode.Type.DIRECTORY).build();

                        try (SFTPClient sftpClient = ssh.newSFTPClient()) {
                            FileAttributes att = sftpClient.statExistence(remotePath);

                            LOG.debug("ExperimentSpecification.dir checked dir {} on {}. exists={}",
                                    remotePath, node.getClientId(), att != null);

                            if (att != null) {
                                //handle existing dir
                                if (att.getType() != FileMode.Type.DIRECTORY) {
                                    throw new IllegalStateException("Error while checking dir "+remotePath+", it already exists but is not a directory");
                                }

                                //don't check permissions, just set them
                                sftpClient.chmod(remotePath, targetDirMask);
                            } else {
                                //create dir(s)
                                SFTPEngine engine = sftpClient.getSFTPEngine();
                                final Deque<String> dirsToMake = new LinkedList<>();
                                for (PathComponents current = engine.getPathHelper().getComponents(remotePath); ;
                                     current = engine.getPathHelper().getComponents(current.getParent())) {
                                    final FileAttributes attrs = sftpClient.statExistence(current.getPath());
                                    if (attrs == null) {
                                        dirsToMake.push(current.getPath());
                                    } else if (attrs.getType() != FileMode.Type.DIRECTORY) {
                                        throw new IllegalStateException("Error while creating dir "+remotePath+" -> "+current.getPath() + " exists but is not a directory");
                                    } else {
                                        break;
                                    }
                                }
                                LOG.debug("ExperimentSpecification.dir must create dirs {}", dirsToMake);
                                while (!dirsToMake.isEmpty()) {
                                    String dirToMake = dirsToMake.pop();
                                    LOG.debug("ExperimentSpecification.dir creating {} on {}",
                                            dirToMake, node.getClientId());
                                    //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work...
                                    engine.makeDir(dirToMake, targetDirAtts);
                                    //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work...
                                    sftpClient.chmod(dirToMake, targetDirMask);
                                }
                            }

                            //TODO: we should set appropriate permissions along the parts of the directory path that already exist. ("execute" perm for uog depending if they get any permissions on final dir)

                            eSpecLogger.firePostDir(dirSpec, remotePath, clientId, true);
                        } catch (IOException ex) {
                            LOG.error("Error while creating dir {} on {}", remotePath, node.getClientId(), ex);
                            updateMessage(String.format("Error while creating dir %s on %s: %s", remotePath, node.getClientId(), ex.getMessage()));
                            eSpecLogger.firePostDir(dirSpec, remotePath, clientId, false);
                            return ExperimentTaskStatus.FAILED;
                        }
                    }
                }

                return ExperimentTaskStatus.SUCCESS;
            }
        }

        public class UploadState extends State {
            protected UploadState() {
                super("Uploading files on " + node.getClientId());
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                if (ssh == null)
                    throw new IllegalStateException("ssh == null");
                if (externalFileUtil == null)
                    throw new IllegalStateException("externalFileUtil == null");

                //We need to wait for all files to be ready
                //noinspection ConstantConditions
                job.getExperiment().getExperimentSpecificationFileManager().waitUntilReady();

                //uploads file AND scripts (but don't execute scripts yet)
                List<UploadLikeSpec> uploads = new ArrayList<>(experimentSpecification.getUploads());
                uploads.addAll(experimentSpecification.getExecutes());

                String uploadDirSpecPath = null;
                String scriptDirSpecPath = null;
                for (DirSpec dirSpec : experimentSpecification.getDirs()) {
                    if (dirSpec.getNodes() == null || dirSpec.getNodes().contains(clientId)) {
                        if (dirSpec.isScriptContent()) {
                            scriptDirSpecPath = dirPathToAbsolutePath(dirSpec.getPath());
                        }
                        if (dirSpec.isUploadContent()) {
                            uploadDirSpecPath = dirPathToAbsolutePath(dirSpec.getPath());
                        }
                    }
                }

                try (SFTPClient sftpClient = ssh.newSFTPClient()) {

                    for (UploadLikeSpec uploadSpec : uploads) {
                        if (uploadSpec.getNodes() == null || uploadSpec.getNodes().contains(clientId)) {
                            byte[] content = job.getExperiment().getExperimentSpecificationFileManager().getBytes(uploadSpec.getSource());

                            assert externalFileUtil != null;
                            //use dir with "scripts" or "uploads" type as appropriate, fall back to homedir
                            String remotePath;
                            if (uploadSpec.getPath() == null) {
                                if (uploadSpec instanceof UploadSpec) {
                                    if (uploadDirSpecPath != null) {
                                        remotePath = assureEndWithSlash(uploadDirSpecPath) + uploadSpec.getSource().getBasename();
                                    } else {
                                        remotePath = assureEndWithSlash(externalFileUtil.getHomeDir()) + uploadSpec.getSource().getBasename();
                                    }
                                } else {
                                    if (uploadSpec instanceof ExecuteSpec) {
                                        if (scriptDirSpecPath != null) {
                                            remotePath = assureEndWithSlash(scriptDirSpecPath) + uploadSpec.getSource().getBasename();
                                        } else {
                                            remotePath = assureEndWithSlash(externalFileUtil.getHomeDir()) + uploadSpec.getSource().getBasename();
                                        }
                                    } else {
                                        throw new RuntimeException("BUG: uploadSpec of unknown class " + uploadSpec.getClass().getName());
                                    }
                                }
                            } else {
                                remotePath = uploadSpec.getPath();
                            }

                            String name = externalFileUtil.getFilename(remotePath);

                            int targetFileMask = ChModPermissionsUtil.toMask(uploadSpec.getPermissions());

                            try {
                                LOG.debug("ExperimentSpecification.upload is uploading {} ({} bytes) to {} on {}",
                                        name, content.length, remotePath, node.getClientId());
                                //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work...
                                sftpClient.getFileTransfer().upload(
                                        new ByteArraySshjFile(name, content, uploadSpec.getPermissions()),
                                        remotePath);
                                //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work...
                                sftpClient.chmod(remotePath, targetFileMask);

                                remotePathByUploadLikeSpec.put(uploadSpec, remotePath);

                                eSpecLogger.firePostUpload(uploadSpec, remotePath, clientId, true);
                            } catch (IOException ex) {
                                LOG.error("Error while uploading {} to {} on {}", name, remotePath, node.getClientId(), ex);
                                updateMessage(String.format("Error while uploading %s to %s on %s: %s", name, remotePath, node.getClientId(), ex.getMessage()));
                                eSpecLogger.firePostUpload(uploadSpec, remotePath, clientId, false);
                                return ExperimentTaskStatus.FAILED;
                            }
                        }
                    }
                } catch (IOException e) {
                    throw new JFedException("Failed to create SSH SFTP client", e);
                }

                return ExperimentTaskStatus.SUCCESS;
            }
        }
    }

    @Nonnull
    private static String assureEndWithSlash(@Nonnull String p) {
        if (p.endsWith("/")) return p;
        return p + "/";
    }

    @Nonnull
    private static String assureStartWithSlash(@Nonnull String p) {
        if (p.startsWith("/")) return p;
        return "/" + p;
    }
}
